
module npl30d(
    input NRESET,   // Board hardware reset, low active
    input [15:0]A,  // MCU address bus
    inout [7:0]D,   // MCU data bus
    input NWE,      // MCU write strobe, low active
    input NOE,      // MCU read strobe, low active
    input CS4,      // MCU P31/A23/CS4
    output P9,      // MCU IRQ0
    input P14,      // MCU TXD0
    output P15,     // MCU RXD0
    input P21,      // MCU TXD1
    output P22,     // MCU RXD1
    input P26,      // MCU chip select 2, low active

    output ULED,    // User LED, low active

    output CFON,    // RS232 interface force on
    output CNFOFF,  // RS232 interface force off
    input CNINVAL,  // RS232 interface invalid signal
    input CRD,      // RS232 interface receive data
    output CTD,     // RS232 interface transmit data
    input CCTS,     // RS232 interface clear to send
    output CRTS,    // RS232 interface ready to send
    input CDSR,     // RS232 interface data set ready
    output CDTR,    // RS232 interface data terminal ready
    input CDCD,     // RS232 interface carrier detect
    input CRI,      // RS232 interface ring indicator

    input MCD,      // MMC detect
    output MCLK,    // MMC host to card clock
    output MCMD,    // MMC host to card data
    input MDAT0,    // MMC card to host data
    output MDAT3,   // MMC host to card select
    
    input ALARM,    // Alarm from RTC
    input WAKEUP,   // Wakeup from LAN
    input FBUSY,    // Busy from FLASH
    
    output [15:0]CAR,   // Expansion port address bus
    inout [7:0]CDR,     // Expansion port data bus
    output NCRD,        // Expansion port read strobe
    output NCWR,        // Expansion port write strobe
    output PNCS,        // Expansion port panel select

    input PLCLK1        // PLL clock
);

    parameter RSCR_ADDR = 8'h00;    // RS232 command register address
    parameter IMRL_ADDR = 8'h04;    // Interrupt mask register low byte address
    parameter IMRH_ADDR = 8'h05;    // Interrupt mask register high byte address
    parameter SLRL_ADDR = 8'h0C;    // Signal latch register low byte address
    parameter SLRH_ADDR = 8'h0D;    // Signal latch register high byte address
    parameter SCRL_ADDR = 8'h10;    // Signal clear register low byte address
    parameter SCRH_ADDR = 8'h11;    // Signal clear register high byte address
    parameter MMCDR_ADDR = 8'h14;   // MMC data register address
    parameter XER_ADDR = 8'h18;     // External enable register address
    parameter VIDR_ADDR = 8'h1C;    // Version identifier register address

    reg [7:0]rsc_reg;   // RS232 control register
    reg [7:0]im_reg_l;  // Interrupt mask register
    reg [3:0]im_reg_h;  // Interrupt mask register
    reg [7:0]sl_reg_l;  // Signal latch register
    reg [3:0]sl_reg_h;  // Signal latch register
    reg [2:0]xe_reg;    // External enable register
    reg [7:0]do_reg;    // Data bus output register
    reg [7:0]mmc_sreg;  // MMC shift data register
    reg [3:0]mmc_creg;  // MMC shift count register
    reg mmc_doreg;      // MMC data output register
    reg mmc_coreg;      // MMC clock output register
    reg mmc_ireg;       // MMC interrupt flag
//    reg MMC_CLK_DIS;

//    wire mmc_irq;
    wire mmc_nwe;       // MMC write enable
  
    wire [6:0]RESET0_irq;
    wire [3:0]RESET1_irq;

    assign RESET1_irq = clear1_irq(A, D, P26, NWE, NRESET);

    function [3:0]clear1_irq;
        input [15:0]A;
        input [3:0]D;
        input P26;
        input NWE;
        input NRESET;

        if(NRESET == 0) begin
            clear1_irq = 4'b0000;
        end            
        else if((A == SCRH_ADDR) & (P26 == 0) & (NWE == 1)) begin
            clear1_irq = ~D;
        end            
        else begin
            clear1_irq = 4'b1111;
        end            
    endfunction

    always @(posedge RESET1_irq[0] or posedge CNINVAL) begin
        if(CNINVAL == 1) begin
            sl_reg_h[0] = 1;
        end
        else begin
            sl_reg_h[0] = 0;
        end
    end

    always @(posedge RESET1_irq[1] or negedge CNINVAL) begin
        if(CNINVAL == 0)
            sl_reg_h[1] = 1;
        else
            sl_reg_h[1] = 0;
    end

    always @(posedge RESET1_irq[2] or negedge MCD) begin
        if(MCD == 0)
            sl_reg_h[2] = 1;
        else
            sl_reg_h[2] = 0;
    end

    always @(posedge RESET1_irq[3] or posedge MCD) begin
        if(MCD == 1)
            sl_reg_h[3] = 1;
        else
            sl_reg_h[3] = 0;
    end

    assign RESET0_irq = clear0_irq(A, D, P26, NWE, NRESET);

    function [6:0]clear0_irq;
        input [15:0]A;
        input [6:0]D;
        input P26;
        input NWE;
        input NRESET;

        if(NRESET == 0)
            clear0_irq = 7'b0000000;
        else if((A == SCRL_ADDR) & (P26 == 0) & (NWE == 1))
            clear0_irq = ~D;
        else
            clear0_irq = 7'b1111111;
    endfunction

    always @(posedge RESET0_irq[0] or negedge CCTS) begin
        if(CCTS == 0)
            sl_reg_l[0] = 1;
        else
            sl_reg_l[0] = 0;
    end

    always @(posedge RESET0_irq[1] or negedge CDSR) begin
        if(CDSR == 0)
            sl_reg_l[1] = 1;
        else
            sl_reg_l[1] = 0;
    end

    always @(posedge RESET0_irq[2] or negedge CDCD) begin
        if(CDCD == 0)
            sl_reg_l[2] = 1;
        else
            sl_reg_l[2] = 0;
    end

    always @(posedge RESET0_irq[3] or posedge CRI) begin
        if(CRI == 1)
            sl_reg_l[3] = 1;
        else
            sl_reg_l[3] = 0;
    end

    always @(posedge RESET0_irq[4] or negedge ALARM) begin
        if(ALARM == 0)
            sl_reg_l[4] = 1;
        else
            sl_reg_l[4] = 0;
    end

    always @(posedge RESET0_irq[5] or posedge WAKEUP) begin
        if(WAKEUP == 1)
            sl_reg_l[5] = 1;
        else
            sl_reg_l[5] = 0;
    end

    always @(posedge RESET0_irq[6] or negedge FBUSY) begin
        if(FBUSY == 0)
            sl_reg_l[6] = 1;
        else
            sl_reg_l[6] = 0;
    end

    assign P9  = ~((|(sl_reg_l & im_reg_l)) | (|(sl_reg_h & im_reg_h)));

    wire NWCS2;
    wire NOCS2;

    assign PNCS = xe_reg[1];
    assign ULED = xe_reg[2];

    assign NWCS2 = ~NWE&CS4;
    assign NOCS2 = ~NOE&CS4;

    assign CAR = CS4 ? A : 16'bzzzzzzzzzzzzzzzz;
    assign NCRD = CS4 ? NOE : 1'bz;
    assign NCWR = CS4 ? NWE : 1'bz;
    assign CDR = NWCS2 ? D : 8'bzzzzzzzz;
    assign D   = NOCS2 ? CDR : 8'bzzzzzzzz;

    //assign MCMD = reg_MMC[0]  ? reg_MMC[2] : 1'bz;
    //assign MDAT0 = reg_MMC[1] ? reg_MMC[3] : 1'bz;

    assign MDAT3 = xe_reg[0];
    

    assign NOECS = ~NOE & ~P26;

    always @(posedge NWE or negedge NRESET) begin
        if(NRESET == 0) begin
            rsc_reg <= 8'b00100000;
            im_reg_l <= 8'b0;
            im_reg_h <= 4'b0;
            xe_reg <= 3'b111;
        end
        else if ((A == RSCR_ADDR) & (P26 == 0)) begin
            rsc_reg <= D;
        end
        else if((A == IMRL_ADDR) & (P26 == 0)) begin
            im_reg_l <= D;
        end
        else if((A == IMRH_ADDR) & (P26 == 0)) begin
            im_reg_h <= D[3:0];
        end
        else if((A == XER_ADDR) & (P26 == 0)) begin
            xe_reg <= D;
        end
    end

    always @(negedge NOE) begin
        do_reg[7:0] = 0;
        if((A == RSCR_ADDR) & (P26 == 0))
            do_reg = rsc_reg;
        if((A == IMRL_ADDR) & (P26 == 0))
            do_reg = im_reg_l;
        if((A == IMRH_ADDR) & (P26 == 0))
            do_reg = im_reg_h;
        if((A == MMCDR_ADDR) & (P26 == 0))
            do_reg = mmc_sreg;
        if((A == XER_ADDR) & (P26 == 0))
            do_reg = xe_reg;
        if((A == SLRL_ADDR) & (P26 == 0))
            do_reg = {mmc_ireg, sl_reg_l[6:0]};
        if((A == SLRH_ADDR) & (P26 == 0))
            do_reg = sl_reg_h;
        if((A == VIDR_ADDR) & (P26 == 0))
            do_reg = 8'h06;
    end

    assign D = NOECS ? do_reg : 8'bzzzzzzzz;




    wire CDTR_buf;

    assign CNFOFF = ~rsc_reg[1];
    assign CFON = rsc_reg[0];
    //assign CNFOFF = 1;
    //assign CFON = 1;

    assign CRTS = ~rsc_reg[3];
    assign CDTR = (( rsc_reg[7] & rsc_reg[5])|(~rsc_reg[7] & rsc_reg[6])) ? CDTR_buf : ~rsc_reg[2];

    // UART0
    assign P15 =  (~rsc_reg[7] & rsc_reg[5]) ? CRD : 1'bz;
    assign P15 =  ( rsc_reg[7] & rsc_reg[5]) ? CDSR : 1'bz;
    assign CTD =  (~rsc_reg[7] & rsc_reg[5]) ? P14 : 1'bz;
    assign CDTR_buf = ( rsc_reg[7] & rsc_reg[5]) ? P14 : P21;

    // UART1
    assign P22  = ( rsc_reg[7] & rsc_reg[6]) ? CRD : 1'bz;
    assign P22  = (~rsc_reg[7] & rsc_reg[6]) ? CDSR : 1'bz;
    assign CTD  = ( rsc_reg[7] & rsc_reg[6]) ? P21 : 1'bz;
    //assign CDTR_buf = (~rsc_reg[7] & rsc_reg[6]) ? P21 : 1'bz;



    assign mmc_nwe = CheckMmcWrite(A, P26, NWE, NRESET);

    function CheckMmcWrite;
        input [15:0]A;
        input P26;
        input NWE;
        input NRESET;

        if(NRESET == 0) begin
            CheckMmcWrite = 0;
        end
        else if((A == MMCDR_ADDR) & (P26 == 0) & (NWE == 0)) begin
            CheckMmcWrite = 0;
        end
        else begin
            CheckMmcWrite = 1;
        end
    endfunction

    // Reading MMC
    always @(posedge mmc_coreg or negedge mmc_nwe) begin
        if (mmc_nwe == 0) begin
            // NRESET MMC Register wird auf 0 gesetzt mmc_creg wird auf 8 gesetzt damit der MMC_CLK gestopt wird
            if(NRESET == 0) begin
                mmc_sreg <= 8'h00;
                mmc_creg = 9;
                mmc_ireg = 0;
            end
            // NWE Daten werden in's reg_MMC geschrieben mmc_creg wird auf 0 gesetzt  MMC_CLK wird aktiviert
            else begin
                mmc_sreg <= D;
                mmc_creg = 0;
                mmc_ireg = 0;
            end
        end
        // Daten werden von der MMC in reg_MMC geschrieben
        else if (mmc_coreg == 1) begin
            if (mmc_creg <  8) begin
                mmc_sreg <= {mmc_sreg[6:0], MDAT0};
                mmc_creg = mmc_creg + 1;
                if (mmc_creg == 7) begin
                    mmc_ireg = 1;
                end
            end
        end
    end

    // MMC Clock
    always @(negedge PLCLK1) begin
        if ((mmc_creg <  8) & (mmc_nwe == 1)) begin
            mmc_coreg = mmc_coreg + 1;
        end
        else begin
            mmc_coreg = 1;
        end
    end

    // MMC Schreiben
    always @(negedge mmc_coreg) begin
        if (mmc_creg <  8) begin
            mmc_doreg = mmc_sreg[7];
        end
    end
    
    assign MCMD = mmc_doreg;
    assign MCLK = mmc_coreg;

endmodule

